/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "MidiMapper.h"
#include <sys/soundcard.h>
#include <vector>

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <math.h>
#include <fcntl.h>

#include <linux/awe_voice.h>

#ifndef _MIDIDEVICE_H_
#define _MIDIDEVICE_H_

// the MidiDeviceInfo stores information about the subdevices exposed by each devices.
class MidiDeviceInfo
{
 public:
  int  id;
  char name[256];
  bool external;
};

/** This is a base class for all the midi devices used by the midnight library.
    
    If you wish to add a new device, write a new class that inherits from MidiDevice.
 */
class MidiDevice
{
 private:
  /// As the name says, it stores the tempo value.
  int               _tempo;

 protected:

  /// The MidiMapper is used to translate the instruments from one standard to another.
  MidiMapper*       _mapper;

  /// Stores all the subdevices for the MidiDevice.
  vector<MidiDeviceInfo> _infos;

  /// The tickstart variable is used when going directly to a certain event : this must be set to inform the device of the jump.
  unsigned long     _tickStart;

 public:
  ///
  MidiDevice();
  ///
  virtual ~MidiDevice();
  /// Sets the midi tempo.
  virtual void setTempo(int tempo) { _tempo = tempo;  }
  /// Gets the midi tempo.
  virtual int getTempo() { return _tempo; }

  /// Sets the _tickStart variable. The tickstart variable is used when going directly to a certain event : this must be set to inform the device of the jump.
  virtual void setTickStart(unsigned long ticks) { _tickStart = ticks; }

  /// Converts the number of ticks in ms (adapted from playmidi)
  int ticksToMs(int divisions, int ticks)
    {
      if (divisions>0)
	{
	  double value = ((double)ticks*(double)_tempo) / (double)divisions;
	  return (int) value;
	}
      else
	{
	  return ticks / (((divisions&0xff)>>8)*(divisions&0xff));
	}
    }

  
 protected:
  /// This allows the inherited classes to make their own MidiDeviceInfo structures available.
  void addInfo(MidiDeviceInfo info);

 public:
  /** 
      Returns an iterator on the sequence of MidiDeviceInfo structures of the device.
   */
  vector<MidiDeviceInfo>::iterator getInfoBegin();
  /** 
      Returns the last iterator on the sequence of MidiDeviceInfo
      structures. As usual, this iterator does not contain a valid
      structure.
  */
  vector<MidiDeviceInfo>::iterator getInfoEnd();

  /// Starts the device timer
  virtual void startTimer() { }
  /// Stops the device timer
  virtual void stopTimer() { }
  /** Wait for a given time
      @param divisions is the number of division, as given by the MidiFile.
      @param ticks the number of midi ticks to wait
  */
  virtual void waitUntil(unsigned long divisions, unsigned long ticks) { }
  /// Flushes all the events sent to the device, including the wait events.
  virtual void flush() { }

  /// Sends an aftertouch event.
  virtual void afterTouch(int channel, int note, int velocity) { }

  /// Send a key press event.
  virtual void noteOn(int channel, int note, int velocity) { }

  /// Send a key release event.
  virtual void noteOff(int channel, int note, int velocity) { }

  /// Set the global volume for the device.
  virtual void setVolume(int volume) { }
  /// Send a bender event.
  virtual void sendBender(int channel, int data1) { }
  /// Send a channel after touch event.
  virtual void sendChannelAfterTouch(int channel, int data1) { }
  /// Send a patch change (instrument change) event.
  virtual void sendPatchChange(int channel, int data1) { }
    
  /// Send a "control" event.
  virtual void sendControlChange(int channel, int data1, int data2) { }

  /** Flushes the events, and synchronizes the device.
      This method returns only when all the pending midi events have been played.
  */
  virtual void sync() { }

  /// Shut up everything.
  virtual void terminateAll() { }

}///
;

#endif

