/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "MidiEvent.h"
#include <vector>

void MidiEvent::init(istream &in, int &length, unsigned long offset, int &runningStatus)
{
  _deltaTime = IOUtils::getIntVar(in, length);
  _offset = offset + (unsigned long) _deltaTime;
  int command = IOUtils::getByte(in, length);
  _data = NULL;
  //
  // Test for meta events
  //

  if (command == 0xff)
    {
        command = IOUtils::getByte(in, length);
	int clen = IOUtils::getIntVar(in, length);
	
	switch (command)
	  {
	  case 0x00:
	    {
	      MidiMetaSequenceNumber *mmsn = new MidiMetaSequenceNumber(clen);
	      for (int i=0; i<clen; i++)
		{
		  mmsn->setSequence(i, IOUtils::getByte(in, length));
		}
	      _data = mmsn;
	      _eventCode = META_SEQUENCE_NUMBER;
	    }
	    return;

	    case 0x01:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_TEXT_EVENT;
	    }
	    return;

	    case 0x02:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_COPYRIGHT;
	    }
	    return;

	    case 0x03:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_SEQUENCE_NAME;
	    }
	    return;

	    case 0x04:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_TRACK_INSTRUMENT_NAME;
	    }
	    return;

	    case 0x05:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_LYRIC;
	    }
	    return;

	    case 0x06:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_MARKER;
	    }
	    return;

	  case 0x07:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_CUE_POINT;
	    }
	    return;

	  case 0x2F:
	    {
	      _data = NULL;
	      _eventCode = META_END;
	    }
	    return;

	    case 0x51:
	    {
	      int val = 0;
	      for (int i=0; i<3; i++)
		{
		  val |= (IOUtils::getByte(in,length) << (8*(2-i)));
		}
	      MidiMetaTempo *mmt = new MidiMetaTempo(val);
	      _data = mmt;
	      _eventCode = META_SET_TEMPO;
	    }
	    return;

	  case 0x58:
	    {
	      int data1 = IOUtils::getByte(in,length);
	      int data2 = IOUtils::getByte(in,length);
	      int data3 = IOUtils::getByte(in,length);
	      int data4 = IOUtils::getByte(in,length);
	      _data = new MidiMetaTimeSignature(data1,data2,data3,data4);
	      _eventCode = META_TIME_SIG;
	    }
	    return;
	    
	  case 0x59:
	    {
	      int data1 = IOUtils::getByte(in,length);
	      int data2 = IOUtils::getByte(in,length);
	      _data = new MidiMetaKeySignature(data1,data2);
	      _eventCode = META_KEY_SIG;
	    }
	    return;

	  case 0x7F:
	    {
	      MidiMetaTextEvent *mmte = new MidiMetaTextEvent(clen);
	      char *data = mmte->getDataBuffer();
	      IOUtils::fillBuffer(in, data, clen, length);
	      _data = mmte;
	      _eventCode = META_SPECIFIC;
	    }
	    return;
	  }
	
	_eventCode = META_UNKNOWN;

	MidiGenericEvent *mge = new MidiGenericEvent(command, clen);
	char *data = mge->getBuffer();
	IOUtils::fillBuffer(in,data,clen, length);
	_data = mge;

	return;
    }


  //
  // test for sysex messages
  //

  if ((command == 0xF0) || (command == 0xF7))
    {
      int clen = IOUtils::getIntVar(in, length);      
      MidiGenericEvent *mge = new MidiGenericEvent(command, clen);
      char *data = mge->getBuffer();
      IOUtils::fillBuffer(in,data,clen, length);
      _data = mge;
      _eventCode = SYSEX;
      return;
    }


  //
  // test for midi event commands
  //

  char data1;

  if ((command & 0x80) == 0)
    {
      // running status ...
      data1 = command;
      command = runningStatus;

      if (command == 0)
	{
	  cerr << "ERROR !! NO RUNNING STATUS\n";
	  exit(99);
	}

    }
  else
    {
      runningStatus = command;
      data1 = IOUtils::getByte(in,length);
    }

  

  switch( (command>>4) & 0xf )
    {
    case 0x8: // note off
      {
	char channel, data2;
	channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	data2 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, data2);
	_eventCode = NOTE_OFF;
	//	runningStatus = _eventCode;
      }
      return;

    case 0x9: // note on
      {
	char channel, data2;
	channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	data2 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, data2);
	_eventCode = NOTE_ON;
	//	runningStatus = _eventCode;
      }
      return;

    case 0xA: // after touch
      {
	char channel, data2;
	channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	data2 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, data2);
	_eventCode = KEY_AFTER_TOUCH;
	//	runningStatus = _eventCode;
      }
      return;

    case 0xB: // control change
      {
	char channel, data2;
	channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	data2 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, data2);
	_eventCode = CONTROL_CHANGE;
      }
      return;

    case 0xC: // prog change
      {
	char channel;
	channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, 0);
	_eventCode = PROG_CHANGE;
      }
      return;

    case 0xD: // channel pressure
      {
	char channel;
        channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, 0);
	_eventCode = CHANNEL_AFTER_TOUCH;
      }
      return;

    case 0xE: // pitch wheel
      {
	char channel, data2;
	channel = (command & 0xf);
	//	data1 = IOUtils::getByte(in,length);
	data2 = IOUtils::getByte(in,length);
	_data = new MidiEventCommand( command & 0xf, data1, data2);
	_eventCode = PITCH_WHEEL_CHANGE;
      }
      return;
    }


  //
  // running status...
  // 
  
  switch (runningStatus)
    {
    case 0x8: // note off
      _eventCode = NOTE_OFF;
      break;

    case 0x9: // note on
      _eventCode = NOTE_ON;
      break;

    case 0xA: // after touch
      _eventCode = KEY_AFTER_TOUCH;
      break;

    case 0xB: // control change
      _eventCode = CONTROL_CHANGE;
      break;

    case 0xC: // prog change
      _eventCode = PROG_CHANGE;
      break;

    case 0xD: // channel pressure
      _eventCode = CHANNEL_AFTER_TOUCH;
      break;

    case 0xE: // pitch wheel
      _eventCode = PITCH_WHEEL_CHANGE;
      break;

    default:
      _eventCode = ~0;
    }

  _data = new MidiEventCommand( runningStatus & 0xf , IOUtils::getByte(in,length), IOUtils::getByte(in,length));

  return;
}

ostream &operator << (ostream &o, MidiEvent *me)
{
  switch (me->getEventCode())
    {
    case MidiEvent::META_TEXT_EVENT:
    case MidiEvent::META_COPYRIGHT:
    case MidiEvent::META_SEQUENCE_NAME:
    case MidiEvent::META_TRACK_INSTRUMENT_NAME:
    case MidiEvent::META_LYRIC:
      o << "[" << me->getOffset() << ",text:" << ((MidiMetaTextEvent*)me->getData())->getText() << "]";
      break;

    case MidiEvent::NOTE_ON:
      {
	MidiEventCommand *mec = (MidiEventCommand*) me->getData();
	o.form("NOTEON:ch=%x,note=%x,vel=%x]", mec->getChannel(), mec->getData1(), mec->getData2());
      }
      break;
    case MidiEvent::NOTE_OFF:
      {
	MidiEventCommand *mec = (MidiEventCommand*) me->getData();
	o.form("NOTEOFF:ch=%x,note=%x,vel=%x]", mec->getChannel(), mec->getData1(), mec->getData2());
      }
      break;
    case MidiEvent::KEY_AFTER_TOUCH:
    case MidiEvent::PROG_CHANGE:
    case MidiEvent::CHANNEL_AFTER_TOUCH:
      {
	//	MidiEventCommand *mec = (MidiEventCommand*) me->getData();
	//	o.form("[C:%x,%x,%x]", mec->getChannel(), mec->getData1(), mec->getData2());
      }
      break;
    case MidiEvent::PITCH_WHEEL_CHANGE:
      {
	MidiEventCommand *mec = (MidiEventCommand*) me->getData();
	o.form("[PITCHWHEEL:%x,%x,%x=%x]", mec->getChannel(), mec->getData1(), mec->getData2(), ((mec->getData2()<<7) + mec->getData1()));
      }
      break;
    case MidiEvent::CONTROL_CHANGE:
      {
	MidiEventCommand *mec = (MidiEventCommand*) me->getData();
       	o.form("[CONTROL:channel %d, control %d, value %d]", mec->getChannel(), mec->getData1(), mec->getData2());
      }
      break;
    default:
      o << "[" << me->getOffset() << "," << me->getDelta() << "," << me->getEventCode() << "]";
      break;;
    }

  return o;
}


ostream &operator << (ostream &o, vector<pair<unsigned long, MidiEvent*> > &vec)
{
  o << "VEC[" << endl;
  for (vector<pair<unsigned long, MidiEvent*> >::iterator cur = vec.begin(); cur != vec.end(); cur++)
    {
      o << cur->second << endl;
    }
  o << "]VEC" << endl;
  return o;
}
