/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "MidiPlayer.h"
#include <linux/timex.h>

void MidiPlayer::setFile(MidiFile *file)
{
  _midiFile = file;
  MidiUtils::uniquifyTrack(file, _events);
  _started = true;
  _curEvent = _events.begin();
}


void MidiPlayer::stop()
{
  if (_midiFile == 0)
    return;
  if (_midiFile->isValid()==false)
    return;

  //  _playing = false;
  setPlaying(false);
  _started = false;
  _curEvent = _events.begin();
  if (_device != NULL)
    {
      _device->stopTimer();
      _device->terminateAll();
      _device->setTickStart(0);
    }
}

void MidiPlayer::pause()
{
  if (_midiFile == 0)
    return;
  if (_midiFile->isValid()==false)
    return;

  //  _playing = false;
  setPlaying(false);

  _started = false;

  if (_device != NULL)
    {
      _device->stopTimer();
      _device->terminateAll();
      _device->setTickStart( _curEvent->first);
    }
}

void MidiPlayer::play()
{
  if (_midiFile == 0)
    return;
  if (_midiFile->isValid()==false)
    return;

  if ((isPlaying() == true) || (_curEvent == 0) || (_device == NULL))
    {
      return;
    }
  
  setPossiblyPlaying(true);

  updateMetaEvents(_curEvent->first);

  //  _playing = true;
  setPlaying(true);
  unsigned long curoffset = 0;

  while ((_curEvent != _events.end()) && isPlaying())
    {
      if (_started == false)
	{
	  _device->terminateAll();
	  _device->stopTimer();
	  
	  _started = true;
	  _device->startTimer();
	  //      curoffset = _curEvent->first;
	  curoffset = 0;
	}

      pair<unsigned long, MidiEvent*> p1 = *_curEvent;

      if (p1.first > curoffset)
	{
	  _device->waitUntil(_midiFile->getDivision(), p1.first);

	  _device->flush();	      
	  curoffset = p1.first;
	  _device->sync();
	}
	
      playEvent(p1.second, true);

      _curEvent++;
    }

  /*
  if (_playing == false)
    cerr << "PLAYING==FALSE!" << endl;
  if (_curEvent == _events.end())
    {
      cerr << "PLAYER: END REACHED!" << endl;
    }
  if (_eventEnd == true)
      cerr << "eventEnd == true" << endl;
  else cerr << "eventEnd == true" << endl;
  */

    if ((_eventEnd == true) && (_curEvent == _events.end()))
    {
      eventEnd(_curEvent);
    }
  
  setPlaying(false);
  setPossiblyPlaying(false);
}

void MidiPlayer::playEvent(MidiEvent *event, bool activeEvents)
{
  if (activeEvents && (_eventAny == true))
    eventAny(_curEvent);
  
  switch(event->getEventCode())
    {
    case MidiEvent::NOTE_ON:
      {
	MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	_device->noteOn(mec->getChannel(), mec->getData1(), mec->getData2());
	if (activeEvents && (_eventNoteOn == true))
	  eventNoteOn(_curEvent);
      }
      break;
    case MidiEvent::NOTE_OFF:
      {
	MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	_device->noteOff(mec->getChannel(), mec->getData1(), mec->getData2());
	if (activeEvents && (_eventNoteOff == true))
	  eventNoteOff(_curEvent);
      }
      break;
    case MidiEvent::KEY_AFTER_TOUCH:
      {
	MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	_device->afterTouch(mec->getChannel(), mec->getData1(), mec->getData2());
	if (activeEvents && (_eventKeyAfterTouch == true))
	  eventKeyAfterTouch(_curEvent);
      }
      break;

    case MidiEvent::CONTROL_CHANGE:
      {
	MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	_device->sendControlChange(mec->getChannel(), mec->getData1(), mec->getData2());
	if (activeEvents && (_eventControlChange == true))
	  eventControlChange(_curEvent);
      }
      break;
	           
    case MidiEvent::PROG_CHANGE:
	{
	  MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	  _device->sendPatchChange(mec->getChannel(), mec->getData1());
	  if (activeEvents && (_eventProgChange == true))
	    eventProgChange(_curEvent);
	}
      break;

    case MidiEvent::CHANNEL_AFTER_TOUCH:
      {
	MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	_device->sendChannelAfterTouch(mec->getChannel(), mec->getData1());
	if (activeEvents && (_eventChannelAfterTouch == true))
	  eventChannelAfterTouch(_curEvent);
      }
      break;

    case MidiEvent::PITCH_WHEEL_CHANGE:
      {
	MidiEventCommand *mec = (MidiEventCommand*)event->getData();
	_device->sendBender(mec->getChannel(), (mec->getData2()<<7) + mec->getData1());
	if (activeEvents && (_eventPitchWheel == true))
	  eventPitchWheel(_curEvent);
      }
      break;

    case MidiEvent::META_SET_TEMPO:
      {
	MidiMetaTempo *mmt = (MidiMetaTempo*) event->getData();
	_device->setTempo(mmt->getTempo());
	if (activeEvents && (_eventTempo == true))
	  eventTempo(_curEvent);

      }
      break;

    case MidiEvent::META_TEXT_EVENT:
      {
	if (activeEvents && (_eventText == true))
	  eventText(_curEvent);
      }
      break;

    case MidiEvent::META_LYRIC:
      {
	if (activeEvents && (_eventLyric == true))
	  eventLyric(_curEvent);
      }
      break;

    case MidiEvent::META_TRACK_INSTRUMENT_NAME:
      {
	if (activeEvents && (_eventTrackInstrumentName == true))
	  eventTrackInstrumentName(_curEvent);
      }
      break;
    default:
	  
      break;
    }
  
}

unsigned long MidiPlayer::totalTicks()
{
  if (_events.begin() != _events.end())
    {
      vector<pair<unsigned long, MidiEvent*> >::iterator iter = _events.end();
      return (--iter)->first;
    }
  return 0;
}

void MidiPlayer::setPosition(unsigned long ticks)
{
  if (_midiFile->isValid())
    {
      _curEvent = updateMetaEvents(ticks);
      
      _started = false;
      
      if (_device != NULL)
	_device->setTickStart(_curEvent->first);
    }
}

