/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "MidiDevice.h"

#ifndef _OSSMIDIDEVICE_H_
#define _OSSMIDIDEVICE_H_

/** This class implements a MidiDevice that uses the OSS sound device.
    Only the features common to external and internal synths are coded
    here.
*/
class OSSMidiDevice : public MidiDevice
{
 protected:

#define _seqbuf _eventsBuffer
#define _seqbufptr _eventsBufferOffset
#define OSSBUFFERSIZE 2000
#define _seqbuflen OSSBUFFERSIZE
  ///
  int _eventsBufferOffset;
  ///
  char _eventsBuffer[2048];

  /// the device handle.
  int _seq;
  /// the frequence of the computer. Usually 100Hz on intel, but alpha gives 1024Hz.
  int _frequence;
  /// the ticks are computed for 1MHz, so the frequence ratio variable give the ratio according to _frequence.
  int _frequenceRatio;
  /// how many internal synth devices ?
  int _lastSynthId;
  /// how many external synth devices ?
  int _lastMidiId;

  /// Opening the device may fail : this indicates if the constructor successfully opened the device.
  bool _valid;

  /// 
  unsigned long _curt;
  ///
  unsigned long _prevTick;
  ///
  double        _prevTime;

 public:
  ///
  OSSMidiDevice(const char *devicename)
    {
      _seq = open(devicename, O_WRONLY, 0);
      if (_seq == -1) 
	{
	  //	  exit(0);
	  _valid = false;
	  return;
	}
      _valid = true;

      //
      // Find all the available external synth
      //

      int nbMidi;
      if (ioctl(_seq, SNDCTL_SEQ_NRMIDIS, &nbMidi) == -1)
	exit(9);

      _lastMidiId = 0;

      for (int i=0; i<nbMidi; i++)
	{
	  midi_info minfo;
	  minfo.device = i;
	  if (ioctl(_seq, SNDCTL_MIDI_INFO, &minfo) != -1)
	    {
	      MidiDeviceInfo mdi;
	      strcpy(mdi.name, minfo.name);
	      mdi.id = i;
	      mdi.external = true;

	      addInfo(mdi);
	    }
	}

      //
      //  Find all the available internal synth 
      // 
      
      int nbSynth;
      if (ioctl(_seq, SNDCTL_SEQ_NRSYNTHS, &nbSynth) == -1)
	exit(9);
      _lastSynthId = nbSynth - 1;
      for (int i=0; i<nbSynth; i++)
	{
	  synth_info sinfo;
	  sinfo.device = i;
	  if (ioctl(_seq, SNDCTL_SYNTH_INFO, &sinfo) == -1)
	    exit(9);

	  MidiDeviceInfo mdi;
      
	  strcpy(mdi.name, sinfo.name);
	  mdi.id = i;
	  mdi.external = false;

	  addInfo(mdi);
	}

      //
      // Find the frequence : typically 100 for Intel, but 1024 for alpha
      //

      _frequence = 0;
      if (ioctl(_seq, SNDCTL_SEQ_CTRLRATE, &_frequence)==-1)
	{
	  cerr << "ERROR AT SNDCTL_SEQ_CTRLRATE!\n";
	  exit(-1);
	}
      _frequenceRatio = 1000000 / _frequence;


      //
      // setup the OSS buffer and misc variables
      //

      _eventsBufferOffset = 0;
      _tickStart = 0;

    }

  ///
  bool operator!()
    {
      return _valid;
    }

  ///
  ~OSSMidiDevice()
    {
      if (_seq>=0)
	close(_seq);
    }

  ///
  void seqbuf_dump ()
    {
      if (_seqbufptr)
	if (write (_seq, _seqbuf, _seqbufptr) == -1)
	  {
	    perror ("can't write to sequencer");
	    exit (-1);
	  }
      _seqbufptr = 0;
    }


  ///
  virtual void startTimer() 
    {
      _curt = 0;
      _prevTick = 0;
      _prevTime = 0.0;
      SEQ_START_TIMER();
      SEQ_DUMPBUF();
    }
  ///
  virtual void stopTimer()
    {
      SEQ_STOP_TIMER();
      SEQ_DUMPBUF();
    }
  
  ///
  virtual void waitUntil(unsigned long divisions, unsigned long ticks);

  ///
  virtual void flush()
    {
      seqbuf_dump();
    }

  ///
  virtual void sync()
    {
      flush();
      ioctl(_seq, SNDCTL_SEQ_SYNC);
	    
    }

  ///
  virtual void setTickStart(unsigned long ticks) 
    {
      MidiDevice::setTickStart(ticks);
      startTimer();
    }


}///
;

#endif


