/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#define PROGRAM_VERSION "midnight v1.0"

#include <getopt.h>
#include <iostream>
#include <strstream.h>
#include <fstream.h>
#include "MidiFile.h"
#include "MidiUtils.h"
#include "MidiDevice.h"
#include "MidiPlayer.h"
#include "SynthOSSDevice.h"
#include "AWEDevice.h"
#include "MidnightConfiguration.h"

void printUsage(ostream &out)
{
  out << "Usage : midnight [options or files...]" << endl;
  out << " options :" << endl;
  out << "  --type, -t with an argument to change the type of output" << endl;
  out << "     where types are either oss-int (oss internal synth), awe," << endl;
  out << "     oss-ext (oss external), or guess (let the system best-guess)." << endl;
  out << "  --output, -o with an argument, gives an hint when guessing." << endl;
  out << "     Values are int (internal) or ext (external)." << endl;
  out << "  --device, -d with an argument, gives the device name." << endl;
  out << "  --devnb, -n with an argument, give the device number, or -1 for any." << endl;
  
}

void printIntro(ostream &out)
{
  out << PROGRAM_VERSION << " (c) Rodrigo Reyes 1999" << endl;
  out << "midnight comes with ABSOLUTELY NO WARRANTY; " << endl;
  out << "This is free software, and you are welcome to redistribute it" << endl;
  out << "under certain conditions." << endl;
}

int play(char *filename, MidiDevice *device)
{
  cout << "playing " << filename << endl;
  ifstream in(filename, ios::in);
  if (!in)
    {
      cerr << filename << " not found, skipping." << endl;
      return 0;
    }

  MidiFile mf(in);
  MidiPlayer play(&mf, device);
  play.play();

  return 1;
}

int main(int argc, char **argv)
{
  if (argc <= 1)
    {
      printIntro(cout);
      printUsage(cout);
      exit(0);
    }

  printIntro(cout);
  MidnightConfiguration config( MidnightConfiguration::getDefaultConfFile().c_str() );
  MidiDevice *device = config.lockDevice();

  // config.setDefault();

  static struct option longOptions[] = 
  {
    {"type", required_argument, 0,'t'},
    {"output", required_argument, 0,'o'},
    {"device", required_argument, 0,'d'},
    {"devnb", required_argument, 0,'n'}
  };

  int res = 0;
  int cnt = 0;
  bool opt_changed = false;
  while (res != -1)
    {
      res = getopt_long(argc, argv, "-t:o:d:n:", longOptions, &cnt);
      switch(res)
	{
	case 't':
	    if (strcmp(optarg, "awe")==0)
	      {
		config.setType(config.OSS_AWE_SYNTH);
		opt_changed = true;
	      }
	    else if (strcmp(optarg, "oss-int")==0)
	      {
		config.setType(config.OSS_INTERNAL_SYNTH);		
		opt_changed = true;
	      }
	    else if (strcmp(optarg, "oss-ext")==0)
	      {
		config.setType(config.OSS_EXTERNAL_MIDI);
		opt_changed = true;
	      }
	    else if (strcmp(optarg, "guess")==0)
	      {
		config.setType(config.GUESS_BEST);
		opt_changed = true;
	      }
	  break;
	case 'o':
	    if (strcmp(optarg, "int")==0)
	      {
		config.setOutput(config.SYNTH_OUTPUT);
		opt_changed = true;
	      }
	    else if (strcmp(optarg, "ext")==0)
	      {
		config.setOutput(config.MIDI_OUTPUT);
		opt_changed = true;
	      }
	  break;
	case 'd':
	  config.setDeviceName(optarg);
	  break;
	case 'n':
	  config.setDeviceNb(atoi(optarg));
	  break;
	default:

	  cerr << "loading " << optarg << endl;

	    if (opt_changed == true)
	      {
		opt_changed = false;
		config.save();
		config.unlockDevice(device);
		device = config.lockDevice();
	      }
	    
	    play(optarg,device);

	}
    }

  config.unlockDevice(device);
}

