/*
  Midnight, a C++ midi library to load and play midi files.
  Copyright (C) 1999 Rodrigo Reyes, reyes@chez.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef _MIDNIGHT_CONFIGURATION_H_
#define _MIDNIGHT_CONFIGURATION_H_

#include "MidiDevice.h"
#include "OSSMidiDevice.h"
#include "MidiUtils.h"
#include <map>
#include <vector>
#include <string>
#include <glib.h>
#include <pwd.h>
#include <sys/types.h>

class MidnightConfigurationListener
{
 public:
  virtual void configurationChanged() = 0;
};

class MidnightConfElement
{
  char *_deviceName;
  int   _deviceNumber;
  int   _deviceType;
  int   _output;
  char *_description;

 public:
  
  MidnightConfElement(const char *deviceName, int devnb, int devtype, int output, char *description)
    {
      _deviceName = new char[strlen(deviceName)+1];
      strcpy(_deviceName, deviceName);
      _deviceNumber = devnb;
      _deviceType = devtype;
      _output = output;

      _description = new char[strlen(description)+1];
      strcpy(_description, description);
    }

  ~MidnightConfElement()
    {
      delete _deviceName;
      delete _description;
    }
  
  MidnightConfElement(MidnightConfElement const & conf)
    {
      _deviceName = new char[strlen(conf.getDeviceName())+1];
      strcpy(_deviceName, conf.getDeviceName());
      _deviceNumber = conf.getDeviceNumber();
      _deviceType = conf.getDeviceType();
      _output = conf.getOutput();

      _description = new char[strlen(conf.getDescription())+1];
      strcpy(_description, conf.getDescription());
    }

  const char *getDescription() const { return _description; }
  const char *getDeviceName() const { return _deviceName; }
  int getDeviceNumber() const { return _deviceNumber; }
  int getDeviceType() const { return _deviceType; }
  int getOutput() const { return _output; }

};

class MidnightConfiguration
{
 public:

  static char *_defaultDeviceName;
  enum {
    UNKNOWN_TYPE, GUESS_BEST, OSS_INTERNAL_SYNTH, OSS_AWE_SYNTH, OSS_EXTERNAL_MIDI
  } Type;

  enum {
    SYNTH_OUTPUT, MIDI_OUTPUT
  };

 private:

  char *_confFileName;

  char* _deviceName;
  int   _deviceNumber;
  int   _deviceType;
  int   _output;
  MidiDevice* _curDevice;
  map<MidiDevice*, int> _lockDevice;
  vector<MidnightConfigurationListener*> _listeners;
  map<string, string> _features;

  void internalLockDevice(MidiDevice *dev)
    {
      map<MidiDevice*, int>::iterator ldit = _lockDevice.find(dev);
      int i = 1;
      if (ldit != _lockDevice.end())
	{
	  i = ldit->second + 1;
	}
      _lockDevice[dev] = i;
    }

  void internalUnlockDevice(MidiDevice *dev)
    {
      if (dev == NULL)
	return;

      map<MidiDevice*, int>::iterator ldit = _lockDevice.find(dev);
      if (ldit != _lockDevice.end())
	{
	  int i = ldit->second - 1;
	  if (i == 0)
	    {
	      _lockDevice.erase(dev);
	      delete dev;
	      return;
	    }
	  _lockDevice[dev] = i;
	}      
    }

 public:
  MidnightConfiguration(const char *filename);
  ~MidnightConfiguration();

  static string getDefaultConfFile()
    {
      struct passwd *pw = getpwuid(getuid());
      string result(pw->pw_dir);
      return result + "/.midnightrc";
    }


  MidiDevice* lockDevice();
  MidiDevice* lockDevice(int);
  void unlockDevice(MidiDevice* dev);

  void addMidnightConfigurationListener(MidnightConfigurationListener *listener)
    {
      _listeners.push_back(listener);
    }

  void removeMidnightConfigurationListener(MidnightConfigurationListener *listener)
    {
      vector<MidnightConfigurationListener*>::iterator i = _listeners.begin();
      bool found = false;
      for (; (i != _listeners.end()) && !found; i++)
	{
	  if (*i == listener)
	    found = true;
	}
      if ( found )
	{
	  _listeners.erase(i);
	}
    }

  void callListeners()
    {
      for (vector<MidnightConfigurationListener*>::iterator i = _listeners.begin(); i!= _listeners.end(); i++)
	{
	  (*i)->configurationChanged();
	}
    }

  void setDefault();
  void setDeviceName(const char *name);
  void setType(int type) { _deviceType = type;  callListeners(); }
  void setOutput(int output) { _output = output; callListeners(); }
  void setDeviceNb(int nb) { _deviceNumber = nb; callListeners(); }
  bool save();

  void setFeature(string key, string value);
  string getFeature(string key, string def);

  void setIntFeature(string key, int value);
  int getIntFeature(string key, int def);

  void configureWith(MidnightConfElement const &element)
    {
      _deviceName = MidiUtils::cppstrdup(element.getDeviceName());
      _output = element.getOutput();
      _deviceNumber = element.getDeviceNumber();
      _deviceType = element.getDeviceType();
      callListeners();
    }

  vector<MidnightConfElement> getElements(const char *deviceName, int &cur)
    {
      vector<MidnightConfElement> elements;
      if (deviceName == NULL)
	deviceName = _deviceName;

      int current = -1;
      cur = -1;
      
      if ( (_curDevice != NULL) && (strcmp(deviceName, _deviceName)==0))
	{
	  for (vector<MidiDeviceInfo>::iterator i=_curDevice->getInfoBegin(); i != _curDevice->getInfoEnd(); i++)
	    {
	      current++;
	      MidiDeviceInfo &curinfo = *i;
	      MidnightConfElement el(deviceName, curinfo.id, GUESS_BEST, (curinfo.external==true)?MIDI_OUTPUT:SYNTH_OUTPUT,
				     curinfo.name);

	      if ( (strcmp(el.getDeviceName(), _deviceName)==0) && (el.getOutput() == _output)
		   && (el.getDeviceNumber() == _deviceNumber))
		cur = current;

	      elements.push_back(el);
	    }
	}
      else
	{
	  OSSMidiDevice device(deviceName);

	  for (vector<MidiDeviceInfo>::iterator i=device.getInfoBegin(); i != device.getInfoEnd(); i++)
	    {
	      current++;
	      MidiDeviceInfo &curinfo = *i;
	      MidnightConfElement el(deviceName, curinfo.id, GUESS_BEST, (curinfo.external==true)?MIDI_OUTPUT:SYNTH_OUTPUT,
				     curinfo.name);

	      if ( (strcmp(el.getDeviceName(), _deviceName)==0) && (el.getOutput() == _output)
		   && (el.getDeviceNumber() == _deviceNumber))
		cur = current;

	      elements.push_back(el);
	    }
	}
      return elements;
    }
  
 private:

  int guessBestType(char *dev, int nb);
  bool load();

};

#endif
